# Copyright 2014 The Chromium Authors. All rights reserved.
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

import os
import sys
import time
import collections

from cStringIO import StringIO

from expect_tests.type_definitions import Handler, Failure
from expect_tests.serialize import GetCurrentData, DiffData, NonExistent


Missing = collections.namedtuple('Missing', 'test log_lines')
Fail = collections.namedtuple('Fail', 'test diff log_lines took_secs')
Pass = collections.namedtuple('Pass', 'test took_secs')
Slow = collections.namedtuple('Slow', 'test took_secs')


class TestHandler(Handler):
  """Run the tests."""
  @classmethod
  def add_options(cls, parser):
    parser.add_argument(
        '--fail-if-longer-than', type=float, default=0,
        help='treat as failures tests that passed but ran longer than given '
             'time (in seconds)')

    parser.add_argument(
        '--emit-tests-per-row', type=int, default=2**30,  # Effectively, unlim.
        help='in non-verbose mode, limit the number of test outputs console '
             'per row.')

  @classmethod
  def run_stage_loop(cls, opts, results, put_next_stage):
    for test, result, log_lines in results:
      current, _ = GetCurrentData(test)
      if current is NonExistent and result.data is not None:
        put_next_stage(Missing(test, log_lines))
      else:
        diff = DiffData(test, current, result.data)
        if diff:
          put_next_stage(Fail(test, diff, log_lines, result.took_secs))
        elif (opts.fail_if_longer_than and
              opts.fail_if_longer_than < result.took_secs):
          put_next_stage(Slow(test, result.took_secs))
        else:
          put_next_stage(Pass(test, result.took_secs))

  class ResultStageHandler(Handler.ResultStageHandler):
    def __init__(self, *args):
      super(TestHandler.ResultStageHandler, self).__init__(*args)
      self.err_out = StringIO()
      self.start = time.time()
      self.errors = collections.defaultdict(int)
      self.num_tests = 0
      self.total_tests_duration_s = 0.0
      self.emit_counter = 0
      self.crash_detected = False
      self.tests_begun = {}

    def _emit(self, short, test, verbose):
      if self.opts.verbose:
        print >> sys.stdout, '%s ... %s' % (test.name if test else '????',
                                            verbose)
      else:
        sys.stdout.write(short)
        self.emit_counter += len(short)
        if self.opts.emit_tests_per_row <= self.emit_counter:
          sys.stdout.write(os.linesep)
          self.emit_counter = 0
        sys.stdout.flush()

    def _add_result(self, msg, test, header, category, log_lines=()):
      print >> self.err_out
      print >> self.err_out, '=' * 70
      if test is not None:
        self._mark_finished(test)
        print >> self.err_out, '%s: %s (%s)' % (
            header, test.name, test.expect_path())
      print >> self.err_out, '-' * 70
      if msg:
        print >> self.err_out, msg
      if log_lines:
        print >> self.err_out, '==== captured logging output ===='
        print >> self.err_out, '\n'.join(log_lines)
      self.errors[category] += 1

    def _mark_finished(self, test):
      assert test.name in self.tests_begun, (
        '%s test not started but finished!' % test.name)
      self.tests_begun.pop(test.name)

    def handle_Pass(self, p):
      self._mark_finished(p.test)
      if not self.opts.quiet:
        self._emit('.', p.test, 'ok')
      self.num_tests += 1
      self.total_tests_duration_s += p.took_secs

    def handle_Slow(self, p):
      self._emit('S', p.test, 'slow')
      self._add_result('test was too slow %.3fs' % p.took_secs, p.test,
                       'SLOW', 'slow tests', [])
      self.num_tests += 1
      self.total_tests_duration_s += p.took_secs

    def handle_Fail(self, fail):
      self._emit('F', fail.test, 'FAIL')
      self._add_result('\n'.join(fail.diff), fail.test, 'FAIL', 'failures',
                       fail.log_lines)
      self.total_tests_duration_s += fail.took_secs
      self.num_tests += 1
      return Failure()

    def handle_TestSkipped(self, skipped):
      self._mark_finished(skipped.test)
      self._emit('S', skipped.test, 'SKIPPED')
      self.num_tests += 1

    def handle_TestError(self, test_error):
      self._emit('E', test_error.test, 'ERROR')
      self._add_result(test_error.message, test_error.test, 'ERROR', 'errors',
                       test_error.log_lines)
      self.num_tests += 1
      return Failure()

    def handle_UnknownError(self, error):
      self._emit('U', None, 'UNKNOWN ERROR')
      self._add_result(error.message, None, 'UNKNOWN ERROR', 'unknown_errors')
      return Failure()

    def handle_CrashedProcessError(self, error):
      if not self.crash_detected:
        self.crash_detected = True
        self._add_result(error.message, None, 'PROCESS CRASHED ERROR',
                         'process_crashed_errors')
      return Failure()

    def handle_PreResult(self, pre_result):
      self.tests_begun[pre_result.test.name] = pre_result.test

    def handle_Missing(self, missing):
      self._emit('M', missing.test, 'MISSING')
      self._add_result('', missing.test, 'MISSING', 'missing',
                       missing.log_lines)
      return Failure()

    def finalize(self, aborted):
      if not aborted:
        # When aborted, many tests won't be finished.
        for test in self.tests_begun.values():
          self._emit('C', test, 'CRASHED')
          self._add_result('', test, 'CRASHED', 'crashed')

      # TODO(iannucci): print summary stats (and timing info?)
      buf = self.err_out.getvalue()
      if buf:
        print
        print buf
      if not self.opts.quiet:
        print
        print '-' * 70
        print 'Ran %d tests in %0.3fs (actual tests runtime: %0.3fs)' % (
            self.num_tests, time.time() - self.start, self.total_tests_duration_s)
        print
      if aborted:
        print 'ABORTED'
      elif self.errors:
        print 'FAILED (%s)' % (', '.join('%s=%d' % i
                                         for i in self.errors.iteritems()))
      elif not self.opts.quiet:
        print 'OK'
